class Ecohab5Gui extends Gui {
  
  Toolbar controller;
  TextButton surfBtn, deepBtn;
  Slider speedSlider;
  TextLabel dateLabel, instructionsLabel;
  
  Ecohab5Gui() {
    controller = new Toolbar(0, styles.guiMargin, width, styles);
    dateLabel = controller.addTextLabel("", 110);
    dateLabel.textSize = styles.largeFontSize;
    surfBtn = controller.addTextButton("hide surface flow");
    surfBtn.outline = true;
    deepBtn = controller.addTextButton("hide deep flow");
    deepBtn.outline = true;
    controller.addSpacer();
    controller.addSpacer();
    speedSlider = controller.addSlider("animation \nspeed", 4*styles.guiHeight, -300, 0);
    speedSlider.showValOnRight = false;
    controller.addSpacer();
    controller.addSpacer();
    instructionsLabel = controller.addTextLabel("");
  }
  
  boolean over() {
    return mouseY <= styles.guiHeight + 2*styles.guiMargin;
  }
  
  void update() {
    if (display==null) return;
    fill(200,180);
    stroke(150);
    strokeWeight(1);
    rect(0, 0, width, styles.guiHeight + 2*styles.guiMargin);
    dateLabel.name = "Sep " + round(display.currTime / 86400. + 2.885) + " 2005";
    if (!display.animating) {
      instructionsLabel.name = "14 d tracks. Click and drag on the map to animate.";
    } else {
      instructionsLabel.name = "Click the map to stop the animation.";
    }
    boolean hit = controller.update();
    if (hit) {
      if (controller.lastUpdated.equals("hide surface flow")) {
        surfBtn.name = "show surface flow";
        display.flows[1].visible = false;
      } else if (controller.lastUpdated.equals("show surface flow")) {
        surfBtn.name = "hide surface flow";
        display.flows[1].visible = true;      
      } else if (controller.lastUpdated.equals("hide deep flow")) {
        deepBtn.name = "show deep flow";
        display.flows[0].visible = false;
      } else if (controller.lastUpdated.equals("show deep flow")) {
        deepBtn.name = "hide deep flow";
        display.flows[0].visible = true;    
      }
    }
    display.animFrameInterval = abs(speedSlider.getVal());
  }

  boolean offerMousePress() {
    controller.offerMousePress();
    return over();
  }
  
  boolean offerMouseDrag() {
    return over();
  }
  
  boolean offerMouseRelease() {
    return over();
  }
}



class Ecohab5Display extends Display {
  String basedir = ROOTDIR + "ecohab5/";
  boolean hourly = true;
  
  Ecohab5Display() {
    modelCredits = "Salish Sea circulation model by Parker MacCready, Dave Sutherland, and Neil Banas";
    Flow F,F2;
    String[] files = new String[14*12], files2 = new String[14*12];
    
    for (int i=0; i<files.length; i++) files[i] = basedir + "flowmaps-deep/ecohab5deep." + i + ".map.nc";
    F = new Flow(files, true);
    F.cyclic = false;
    for (int i=0; i<files.length; i++) F.T[i] = 44712/12*2 * (i+0.5);
    addFlow(F, transparency(dkRedColor,0.8));
    
    for (int i=0; i<files2.length; i++) files2[i] = basedir + "flowmaps-surf/ecohab5." + i + ".map.nc";
    F2 = new Flow(files2, true);
    F2.cyclic = false;
    for (int i=0; i<files2.length; i++) F2.T[i] = 44712/12*2 * (i+0.5);
    addFlow(F2, transparency(orangeColor,0.8));
    
    timestep = 44712/12*2;
    hoverPathDuration = 14*(44712*2); // 14 lunar days
    animFrameInterval = 50;
    animEndTime = flows[0].lastTime();
    animParticleReps = 1;

    MapAxes A1 = new MapAxes("fulldomain", basedir + "basemap_126_122_4525_50.jpg", new float[] {0, 0, 455, 800}, new float[] {-126, 45.25, 4, 4.75});
    addAxes(A1);
    
    MapAxes A2 = new MapAxes("pugetsound", basedir + "basemap_1232_1222_47_483.jpg", new float[] {A1.right()+2, 0, 416, 800}, new float[] {-123.2, 47, 1, 1.3});
    addAxes(A2);
    
    A1.addInset(A2);
    
    size(round(A1.wd+A2.wd+2), round(A1.ht));
  }
}
