import processing.core.*; 
import processing.xml.*; 

import ucar.*; 
import java.io.IOException; 
import org.slf4j.LoggerFactory; 

import dods.dap.*; 
import thredds.catalog.parser.jdom.*; 
import ucar.nc2.dt.image.*; 
import ucar.nc2.ncml.*; 
import dods.util.*; 
import ucar.unidata.io.http.*; 
import dods.dap.parser.*; 
import thredds.cataloggen.inserter.*; 
import ucar.nc2.dt.trajectory.*; 
import ucar.nc2.iosp.nexrad2.*; 
import ucar.atd.dorade.*; 
import ucar.nc2.iosp.gini.*; 
import ucar.nc2.thredds.*; 
import ucar.nc2.iosp.nids.*; 
import ucar.nc2.iosp.dorade.*; 
import thredds.util.*; 
import ucar.nc2.util.*; 
import ucar.nc2.dataset.*; 
import thredds.cataloggen.*; 
import thredds.catalog.query.*; 
import ucar.nc2.units.*; 
import thredds.crawlabledataset.sorter.*; 
import thredds.cataloggen.datasetenhancer.*; 
import thredds.crawlabledataset.*; 
import ucar.units.*; 
import org.slf4j.impl.*; 
import ucar.nc2.dt.grid.*; 
import ucar.nc2.dods.*; 
import ucar.unidata.util.*; 
import ucar.nc2.dataset.conv.*; 
import ucar.unidata.io.*; 
import ucar.nc2.geotiff.*; 
import ucar.nc2.dataset.grid.*; 
import org.slf4j.*; 
import ucar.nc2.dt.point.*; 
import ucar.nc2.dt.radial.*; 
import ucar.nc2.adde.*; 
import ucar.ma2.*; 
import org.slf4j.spi.*; 
import thredds.catalog.*; 
import dods.dap.Server.*; 
import thredds.datatype.*; 
import ucar.unidata.geoloc.vertical.*; 
import thredds.crawlabledataset.filter.*; 
import ucar.unidata.geoloc.*; 
import ucar.nc2.iosp.dmsp.*; 
import ucar.nc2.dt.*; 
import ucar.unidata.geoloc.projection.*; 
import ucar.nc2.iosp.grib.*; 
import ucar.nc2.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class flowWeaver extends PApplet {

// flowWeaver 0.4.3
// neil banas, dec 2010

String vizCredits = "flowWeaver 0.4.3 by Neil Banas";
//String ROOTDIR = "/Users/neil/Documents/Processing/work/flowWeaver/"; // when running inside Processing
String ROOTDIR = ""; // when exporting as an app

/*
new:
  simple toolbar with toggles for the surface and deep flow and a speed slider
*/

/*
to do:
  split this into a trajectory server and a client
  be able to use a series of map background images (to display a tracer)
  gui
    time control
      slider attached to an axes
      plot wind and tides in little axes
    speed control
    rewind, play/pause
    more info
    tools: streaklines, particle patches, particle fountains
  fix pathnames for export as a windows app
  make Axes correctly clip lines to the boundaries (as opposed to points, which already work)
  age fountain
  higher-level methods for AnimatedParticleSet and TrajectorySet, to separate the calculation from the user-interaction stuff?
    this would also make it easy to have multiple animated particle sets
*/

Gui gui;
Display display;
Stylesheet masterStylesheet;
boolean debug = true;
boolean userStartedDoingStuff = false;

public void setup() {
  size(873,800);
  smooth();
  if (online) debug = false;
  masterStylesheet = new Stylesheet();
  masterStylesheet.loadFonts();
  display = new Ecohab5Display();
  gui = new Ecohab5Gui();
  gui.display = display;
  display.gui = gui;

}

public void draw() {
  display.update();
  gui.update();
  if (display.preloading()) {
    display.preloadNext();
    display.drawPreloadingStatus();
  }
}

public void mousePressed() {
  gui.offerMousePress();
}

public void mouseDragged() {
  if (!gui.offerMouseDrag()) display.offerMouseDrag();
}

public void mouseReleased() {
  if (!gui.offerMouseRelease()) display.offerMouseRelease();
}

public void mouseMoved() {
  userStartedDoingStuff = true;
}

class Ecohab5Gui extends Gui {
  
  Toolbar controller;
  TextButton surfBtn, deepBtn;
  Slider speedSlider;
  TextLabel dateLabel, instructionsLabel;
  
  Ecohab5Gui() {
    controller = new Toolbar(0, styles.guiMargin, width, styles);
    dateLabel = controller.addTextLabel("", 110);
    dateLabel.textSize = styles.largeFontSize;
    surfBtn = controller.addTextButton("hide surface flow");
    surfBtn.outline = true;
    deepBtn = controller.addTextButton("hide deep flow");
    deepBtn.outline = true;
    controller.addSpacer();
    controller.addSpacer();
    speedSlider = controller.addSlider("animation \nspeed", 4*styles.guiHeight, -300, 0);
    speedSlider.showValOnRight = false;
    controller.addSpacer();
    controller.addSpacer();
    instructionsLabel = controller.addTextLabel("");
  }
  
  public boolean over() {
    return mouseY <= styles.guiHeight + 2*styles.guiMargin;
  }
  
  public void update() {
    if (display==null) return;
    fill(200,180);
    stroke(150);
    strokeWeight(1);
    rect(0, 0, width, styles.guiHeight + 2*styles.guiMargin);
    dateLabel.name = "Sep " + round(display.currTime / 86400.f + 2.885f) + " 2005";
    if (!display.animating) {
      instructionsLabel.name = "14 d tracks. Click and drag on the map to animate.";
    } else {
      instructionsLabel.name = "Click the map to stop the animation.";
    }
    boolean hit = controller.update();
    if (hit) {
      if (controller.lastUpdated.equals("hide surface flow")) {
        surfBtn.name = "show surface flow";
        display.flows[1].visible = false;
      } else if (controller.lastUpdated.equals("show surface flow")) {
        surfBtn.name = "hide surface flow";
        display.flows[1].visible = true;      
      } else if (controller.lastUpdated.equals("hide deep flow")) {
        deepBtn.name = "show deep flow";
        display.flows[0].visible = false;
      } else if (controller.lastUpdated.equals("show deep flow")) {
        deepBtn.name = "hide deep flow";
        display.flows[0].visible = true;    
      }
    }
    display.animFrameInterval = abs(speedSlider.getVal());
  }

  public boolean offerMousePress() {
    controller.offerMousePress();
    return over();
  }
  
  public boolean offerMouseDrag() {
    return over();
  }
  
  public boolean offerMouseRelease() {
    return over();
  }
}



class Ecohab5Display extends Display {
  String basedir = ROOTDIR + "ecohab5/";
  boolean hourly = true;
  
  Ecohab5Display() {
    modelCredits = "Salish Sea circulation model by Parker MacCready, Dave Sutherland, and Neil Banas";
    Flow F,F2;
    String[] files = new String[14*12], files2 = new String[14*12];
    
    for (int i=0; i<files.length; i++) files[i] = basedir + "flowmaps-deep/ecohab5deep." + i + ".map.nc";
    F = new Flow(files, true);
    F.cyclic = false;
    for (int i=0; i<files.length; i++) F.T[i] = 44712/12*2 * (i+0.5f);
    addFlow(F, transparency(dkRedColor,0.8f));
    
    for (int i=0; i<files2.length; i++) files2[i] = basedir + "flowmaps-surf/ecohab5." + i + ".map.nc";
    F2 = new Flow(files2, true);
    F2.cyclic = false;
    for (int i=0; i<files2.length; i++) F2.T[i] = 44712/12*2 * (i+0.5f);
    addFlow(F2, transparency(orangeColor,0.8f));
    
    timestep = 44712/12*2;
    hoverPathDuration = 14*(44712*2); // 14 lunar days
    animFrameInterval = 50;
    animEndTime = flows[0].lastTime();
    animParticleReps = 1;

    MapAxes A1 = new MapAxes("fulldomain", basedir + "basemap_126_122_4525_50.jpg", new float[] {0, 0, 455, 800}, new float[] {-126, 45.25f, 4, 4.75f});
    addAxes(A1);
    
    MapAxes A2 = new MapAxes("pugetsound", basedir + "basemap_1232_1222_47_483.jpg", new float[] {A1.right()+2, 0, 416, 800}, new float[] {-123.2f, 47, 1, 1.3f});
    addAxes(A2);
    
    A1.addInset(A2);
    
    size(round(A1.wd+A2.wd+2), round(A1.ht));
  }
}
class Axes {

  // private
  float x0,y0,wd,ht; // screen location
  float xdmin=0, xdmax=1, ydmin=0, ydmax=1; // data ranges
  String topLabel, bottomLabel, leftLabel, rightLabel;
  boolean largeTopLabel = false;
  String[] topTickLabels, bottomTickLabels, leftTickLabels, rightTickLabels;
  float[] topTicks, bottomTicks, leftTicks, rightTicks;
  float leftLabelOffset=0, rightLabelOffset=0, topLabelOffset=0, bottomLabelOffset=0;
  int col = color(0);
  float fontSize = 12;
  Stylesheet styles = masterStylesheet;
  
  // public
  float[] horizGrid, vertGrid;
  boolean gridOn = true, gridOnTop = true; // on top = after the contents of the plot are drawn rather than before
  boolean axesOn = true;
  boolean labelsOn = true, labelsOnTop = true;
  int bkgndColor = color(255,0);
  int gridColor = color(255);
  
  Axes() {}
  
  Axes(float[] rect) {
    setPosition(rect[0], rect[1], rect[2], rect[3]);
  }
  
  Axes(float x0, float y0, float wd, float ht) {
    setPosition(x0,y0,wd,ht);
  }
  
  public void setPosition(float x0, float y0, float wd, float ht) {
    this.x0 = x0;
    this.y0 = y0;
    this.wd = wd;
    this.ht = ht;
  }
  
  public void setAppearance(int col, float fontSize) {
    this.col = col;
    this.fontSize = fontSize;
  }

  public void setXRange(float xdmin, float xdmax) {
    this.xdmin = xdmin;
    this.xdmax = xdmax;
  }
  
  public void setYRange(float ydmin, float ydmax) {
    this.ydmin = ydmin;
    this.ydmax = ydmax;
  }
  
  public float dataAspectRatio() {
    return ((ydmax-ydmin) / ht) / ((xdmax-xdmin) / wd);
  }

  public void setLabels(String where, String label) {
    if (where.equals("left") || where.equals("left of origin") || where.equals("right") || where.equals("right of origin")) {
      setLabels(where, label, horizGrid);
    } else {
      setLabels(where, label, vertGrid);
    }
  }

  public void setLabels(String where, String label, float[] ticks) {
    String[] tickLabels = null;
    if (ticks != null) {
      tickLabels = new String[ticks.length];
      for (int i=0; i<ticks.length; i++) tickLabels[i] = nicenum(ticks[i]);
    }
    setLabels(where, label, ticks, tickLabels);
  }
  
  public void setLabels(String where, String label, float[] ticks, String[] tickLabels) {
    if (where.equals("left")) {
      leftLabel = label;
      leftTicks = ticks;
      leftTickLabels = tickLabels;
    } else if (where.equals("right")) {
      rightLabel = label;
      rightTicks = ticks;
      rightTickLabels = tickLabels;
    } else if (where.equals("top")) {
      topLabel = label;
      topTicks = ticks;
      topTickLabels = tickLabels;
    } else if (where.equals("bottom")) {
      bottomLabel = label;
      bottomTicks = ticks;
      bottomTickLabels = tickLabels;
    } else if (where.equals("left of origin")) {
      setLabels("left", label, ticks, tickLabels);
      leftLabelOffset = x2scr(0) - left();
    } else if (where.equals("right of origin")) {
      setLabels("right", label, ticks, tickLabels);
      rightLabelOffset = x2scr(0) - right();
    } else if (where.equals("top of origin")) {
      setLabels("top", label, ticks, tickLabels);
      topLabelOffset = y2scr(0) - top();
    } else if (where.equals("bottom of origin")) {
      setLabels("bottom", label, ticks, tickLabels);
      bottomLabelOffset = y2scr(0) - bottom();
    }
  }
  
  public float left() {return x0;}
  public float right() {return x0+wd;}
  public float top() {return y0;}
  public float bottom() {return y0+ht;}
  
  public float xmin() {return xdmin;}
  public float xmax() {return xdmax;}
  public float ymin() {return ydmin;}
  public float ymax() {return ydmax;}
  
  public float scr2x(float s) {return map(s, left(), right(), xmin(), xmax());}
  public float scr2y(float s) {return map(s, bottom(), top(), ymin(), ymax());}
  public float x2scr(float x) {return map(x, xmin(), xmax(), left(), right());}
  public float y2scr(float y) {return map(y, ymin(), ymax(), bottom(), top());}
  
  public boolean over(float x, float y) {
    return (x >= left() && x <= right() && y >= top() && y <= bottom());
  }
  public boolean over() {
    return over(mouseX,mouseY);
  }

  public void drawAxes() {drawAxes(col,1);}
  public void drawAxes(int col, float wt) {
    pushStyle();
    stroke(col);
    strokeWeight(wt);
    fill(bkgndColor);
    rectMode(CORNERS);
    rect(left(), top(), right(), bottom());
    popStyle();
  }
  
  public void drawLabels() {drawLabels(col,fontSize);}
  
  public void drawLabels(int col, float fontSize) {
    float sp = fontSize/4;
    pushStyle();
    fill(col);
    styles.setFont(fontSize);    
    if (leftLabel != null || leftTickLabels != null) {
      textAlign(RIGHT);
      float x = left() + leftLabelOffset - sp;
      float maxWd = 0;
      if (leftTickLabels != null) {
        for (int i=0; i<leftTickLabels.length; i++) {
          textVAlign(leftTickLabels[i], x, y2scr(leftTicks[i]), "middle");
          maxWd = max(maxWd, textWidth(leftTickLabels[i]));
        }
        x -= maxWd - sp;
      }
      if (leftLabel != null) {
        pushMatrix();
        translate(x, 0.5f*(top()+bottom()));
        rotate(HALF_PI);
        textAlign(CENTER);
        text(leftLabel);
        popMatrix();
      }
    }
    if (rightLabel != null || rightTickLabels != null) {
      textAlign(LEFT);
      float x = right() + rightLabelOffset + sp;
      float maxWd = 0;
      if (rightTickLabels != null) {
        for (int i=0; i<rightTickLabels.length; i++) {
          textVAlign(rightTickLabels[i], x, y2scr(rightTicks[i]), "middle");
          maxWd = max(maxWd, textWidth(rightTickLabels[i]));
        }
        x += maxWd + sp;
      }
      if (rightLabel != null) {
        pushMatrix();
        translate(x, 0.5f*(top()+bottom()));
        rotate(-HALF_PI);
        textAlign(CENTER);
        text(rightLabel);
        popMatrix();
      }
    }
    if (topLabel != null || topTickLabels != null) {
      textAlign(CENTER);
      float y = top() + topLabelOffset - sp;
      if (topTickLabels != null) {
        for (int i=0; i<topTickLabels.length; i++) {
          text(topTickLabels[i], x2scr(topTicks[i]), y);
        }
        y -= (textAscent() + sp);
      }
      if (topLabel != null) {
        titleText(topLabel, 0.5f*(left()+right()), y); // mechanism for extending classes to write exciting titles
      }
    }
    if (bottomLabel != null || bottomTickLabels != null) {
      textAlign(CENTER);
      float y = bottom() + bottomLabelOffset + sp;
      if (bottomTickLabels != null) {
        for (int i=0; i<bottomTickLabels.length; i++) {
          textVAlign(bottomTickLabels[i], x2scr(bottomTicks[i]), y, "top");
        }
        y += (textAscent() + sp);
      }
      if (bottomLabel != null) {
        text(bottomLabel, 0.5f*(left()+right()), y);
      }
    }
    popStyle();
  } // drawLabels
  
  public void titleText(String S, float x, float y) {
    textAlign(CENTER);
    if (largeTopLabel) styles.setFont(styles.largeFontSize); // this breaks the independence of the Axes class from the font handling in this applet
    text(S,x,y);
    styles.setFont(styles.normalFontSize);
  }
  
  public void drawGrid() {drawGrid(gridColor,0.5f);}
  public void drawGrid(int col, float wt) {
    pushStyle();
    stroke(col);
    strokeWeight(wt);
    if (horizGrid != null) {
      for (int i=0; i<horizGrid.length; i++) line(left(), y2scr(horizGrid[i]), right(), y2scr(horizGrid[i]));
    }
    if (vertGrid != null) {
      for (int i=0; i<vertGrid.length; i++) line(x2scr(vertGrid[i]), top(), x2scr(vertGrid[i]), bottom());
    }
    popStyle();
  }

  // -------------- call these as brackets around drawing the content of the plot
  public void drawBase() {
    if (gridOn && !gridOnTop) drawGrid();
    if (labelsOn && !labelsOnTop) drawLabels();
  }
  
  public void drawOnTop() {
    if (gridOn && gridOnTop) drawGrid();
    if (axesOn) drawAxes();
    if (labelsOn && labelsOnTop) drawLabels();
  }
  // ---------------
  
  
  public void plot(float X, float Y, int col, String lineSpec, float sz) {
    plot(new float[] {X}, new float[] {Y}, col, lineSpec, sz);
  }

  public void plot(float[] X, float[] Y, int col, String lineSpec, float sz) {
    if (lineSpec.equals("o-") || lineSpec.equals("-o")) {
      plot(X,Y,col,"-",sz/4);
      plot(X,Y,col,"o",sz);
    } else {
      stroke(col);
      strokeWeight(sz);
      noFill();
      if (lineSpec.equals("o")) {
        beginShape(POINTS);
      } else {
        beginShape();
      }
      for (int i=0; i<X.length; i++) {
        if (over(x2scr(X[i]), y2scr(Y[i]))) { // only plot points within the axes. Won't work for line segments, just points!
          vertex(x2scr(X[i]), y2scr(Y[i]));
        }
      }
      endShape();
    }
  }

  
}


// -----------------------------------------------------------------------------------------------------------
 
 
 
class MapAxes extends Axes {
  
  String tag;
  PImage bkgndImage;
  String bkgndFilename;
  MapAxes[] insets = new MapAxes[0];
  
  MapAxes(String tag, String filename, float[] scrRect, float[] dataRect) { // give both rects as x0, y0, width, height
    this.tag = tag;
    String suffix = fileSuffix(filename);
    if (suffix==null) {
      println("don't know how to load " + filename + ".");
    } else if (suffix.equals("jpg") || suffix.equals("png")) {
      bkgndFilename = filename;
      bkgndImage = loadImage(bkgndFilename);
    } else {
      println("don't know how to load " + filename + ".");      
    }
    setPosition(scrRect[0], scrRect[1], scrRect[2], scrRect[3]);
    setXRange(dataRect[0], dataRect[0]+dataRect[2]);
    setYRange(dataRect[1], dataRect[1]+dataRect[3]);
    // could do more here regarding labels
    axesOn = false;
    gridOn = false;
  }
  
  public void drawBase() {
    image(bkgndImage, left(), top(), right()-left(), bottom()-top());
    for (int i=0; i<insets.length; i++) {
      noFill();
      stroke(60,80);
      strokeWeight(2);
      rectMode(CORNERS);
      rect(x2scr(insets[i].xdmin), y2scr(insets[i].ydmin), x2scr(insets[i].xdmax), y2scr(insets[i].ydmax));
    }
    super.drawBase();
  }
  
  public void addInset(MapAxes ax) {
    insets = (MapAxes[]) append(insets, ax);
  }
  
}
 




class Display {
  MapAxes[] axes = new MapAxes[0];
  Flow[] flows = new Flow[0];
  int[] flowColors = new int[0];
  float timestep = 0;
  Gui gui;
  Stylesheet styles = masterStylesheet;
  String modelCredits = null;

  float currTime = 0;

  boolean showPathOnHover = true;
  float hoverPathDuration = 10*86400;
  int hoverPathReplicates = 10;
  Trajectory[][] hoverPaths;
  float hoverStartMouseX = -1, hoverStartMouseY = -1;

  float selectX0, selectY0, selectR;
  boolean selecting = false;
  MapAxes selectionAxes;
  int selectColor = color(220,150);
  boolean animating = false;
  int animMaxRadius = 50;
  Particle[] animParticleStart;
  Particle[][] animParticles; // animParticles[j][i] is the ith point within the set of current particle positions for flow j
  float animFrameInterval = 150; // ms
  int animParticleReps = 1;
  float animLastFrameTime;
  float animEndTime = Inf;

  Display() {}

  public MapAxes addAxes(MapAxes A) {
    axes = (MapAxes[]) append(axes, A);
    return A;
  }

  public Flow addFlow(Flow F) {
    return addFlow(F,color(0));
  }
  public Flow addFlow(Flow F, int col) {
    flows = (Flow[]) append(flows, F);
    flowColors = (int[]) append(flowColors, col);
    if (timestep == 0) timestep = F.timestep; // set the display timestep from (probably) the first flow added. This can be overridden.
    return F;
  }
  


  public void preload() {
    for (int i=0; i<flows.length; i++) flows[i].preload();
  }
  
  public void preloadNext() {
    for (int i=0; i<flows.length; i++) flows[i].preloadNext();
  }
  
  public void drawPreloadingStatus() {
    int numLoaded = 0, total = 0;
    for (int i=0; i<flows.length; i++) {
      numLoaded += flows[i].nextToPreload;
      total += flows[i].NF;  
    }
    styles.setFont(styles.largeFontSize);
    fill(styles.orangeColor);
    stroke(255);
    textAlign(LEFT);
    text("loading...\n" + round((0.0f+numLoaded) / total * 100) + "%", 10, 50+styles.largeFontSize);
  }
  
  public boolean preloading() {
    boolean result = false;
    for (int i=0; i<flows.length && result==false; i++) if (flows[i].preloading) result = true;
    return result;
  }



  public MapAxes overWhich() {
    return overWhich(mouseX, mouseY);
  }
  public MapAxes overWhich(float x, float y) {
    for (int i=axes.length-1; i>=0; i--) {
      if (axes[i].over(x,y)) return axes[i];
    }
    return null;
  }

  public boolean over(float x, float y) {
    return overWhich(x,y) != null;
  }
  public boolean over() {
    return overWhich() != null;
  }

  public float[] scr2xy(float sx, float sy) {
    MapAxes ax = overWhich(sx,sy);
    if (ax == null) {
      return null;
    } 
    else {
      return new float[] {
        ax.scr2x(sx), ax.scr2y(sy)
      };
    }
  }



  public void update() {
    // construct a faint trajectory leading forward in time from the point the mouse is hovering over--one path per flowfield
    if (showPathOnHover && over() && !selecting && !animating) {
      if (mouseX != hoverStartMouseX || mouseY != hoverStartMouseY) { // new start position: re-initialize the hover paths
        hoverStartMouseX = mouseX;
        hoverStartMouseY = mouseY;
        float[] xyStart = scr2xy(mouseX, mouseY);
        hoverPaths = new Trajectory[flows.length][hoverPathReplicates];
        for (int i=0; i<flows.length; i++) {
          if (flows[i].visible && flows[i].hoverable) {
            for (int k=0; k<hoverPathReplicates; k++) {
              hoverPaths[i][k] = new Trajectory(xyStart[0], xyStart[1], currTime, flows[i]);
            }
          }
        }
      }
      float calcTime = max(20, animFrameInterval) / flows.length / hoverPathReplicates; // all flows together get animFrameInterval (or at least 20 ms) to update their hover paths per frame
      for (int i=0; i<flows.length; i++) { // update the hover paths 
        for (int k=0; k<hoverPathReplicates; k++) {
          if (hoverPaths[i][k] != null) {
            hoverPaths[i][k].brieflyCalcTowardTime(currTime + hoverPathDuration, calcTime);
          }
        }
      }
    }

    if (animating) {
      // update particle positions
      if (millis() >= animLastFrameTime + animFrameInterval) {
        animLastFrameTime = millis();
        for (int j=0; j<flows.length; j++) {
          if (flows[j].visible && flows[j].animatable) {
            for (int i=0; i<animParticles[j].length; i++) {
              animParticles[j][i].calcToTime(currTime + timestep); // assuming that while animating, the advancing particles control currTime; if the user tries to intervene, the animation stops (or something)
            }
          }
        }
        currTime += timestep;
        if (currTime > animEndTime) startAnimation();
      }
    }

    // in each axes, draw the base background image followed by the hover path for each flowfield, followed by the top layer.
    // all plotting within each axes should be done before the background of the next axes is drawn.
    background(255);
    for (int i=0; i<axes.length; i++) {
      axes[i].drawBase();
      if (!animating && !selecting && userStartedDoingStuff) {
        for (int j=0; j<flows.length; j++) {
          for (int k=0; k<hoverPathReplicates; k++) {
            if (hoverPaths[j][k] != null) {
              float[] x = hoverPaths[j][k].x();
              float[] y = hoverPaths[j][k].y();
              for (int m=0; m<x.length-1; m++) {
                int col = transparency(flowColors[j], (1 - (m+0.0f)/(x.length-2)));
                axes[i].plot(new float[] {x[m],x[m+1]}, new float[] {y[m],y[m+1]}, col, "o-", 1);
              }
            }
          }
        }
      }
      if (animating) {
        // draw particle start locations for the first flow only
        for (int k=0; k<animParticleStart.length; k++) {
          axes[i].plot(animParticleStart[k].x, animParticleStart[k].y, transparency(selectColor, 1.f/animParticleReps), "o", 4);
        } 
        // draw current particle positions
        for (int j=0; j<flows.length; j++) {
          if (flows[j].visible && flows[j].animatable) {
            for (int k=0; k<animParticleStart.length; k++) {            
              if (!animParticles[j][k].stuck()) {
                axes[i].plot(animParticles[j][k].x, animParticles[j][k].y, transparency(flowColors[j], 1.f/sqrt(animParticleReps)) , "o", 4);
              }
            }
          }
        }
      }
      axes[i].drawOnTop();
    }

    if (selecting) {
      fill(selectColor);
      noStroke();
      ellipseMode(CENTER);
      ellipse(selectX0, selectY0, 2*selectR, 2*selectR);
    }
  }



  public void startSelection() {
    selecting = true; // true even if the initial click is in a bad location; that's how we know this drag event has already started
    stopAnimation(); // any click that might start a selection ends the animation
    selectionAxes = overWhich();
    if (selectionAxes != null) {
      selectX0 = mouseX;
      selectY0 = mouseY;
      selectR = 0;
    }
  }

  public void offerMouseDrag() {
    if (!selecting) {
      startSelection();
    }
    if (selecting && selectionAxes != null) {
      float x1 = constrain(mouseX, selectionAxes.left(), selectionAxes.right());
      float y1 = constrain(mouseY, selectionAxes.top(), selectionAxes.bottom());
      selectR = dist(x1, y1, selectX0, selectY0);
      selectR = min(selectR, animMaxRadius);
    }
  }

  public void offerMouseRelease() {
    if (!selecting) { // just a click
      if (animating) {
        stopAnimation();
        return;
      } 
      else {
        startSelection(); // if it's good conditions for a selection, treat a click as a zero-radius selection
      }
    }
    selecting = false;
    if (selectionAxes != null) {
      setAnimStartPositions(selectX0, selectY0, selectR);
      startAnimation();
    }
  }

  public void setAnimStartPositions(float sx0, float sy0, float sr) {
    // convert the selection coordinates to data units
    float[] xy0 = scr2xy(sx0,sy0);
    float x1 = overWhich(sx0,sy0).scr2x(sx0+sr);
    float y1 = overWhich(sx0,sy0).scr2y(sy0+sr);
    float x0 = xy0[0];
    float y0 = xy0[1];
    float rx = x1 - xy0[0];
    float ry = y1 - xy0[1];
    // initialize
    float[] xgrid = flows[0].X;
    float[] ygrid = flows[0].Y;
    animParticleStart = new Particle[0];
    // find all grid points within radius r of (x0,y0)
    for (int i=0; i<xgrid.length; i++) {
      if (xgrid[i] >= x0-rx && xgrid[i] <= x0+rx) {
        for (int j=0; j<ygrid.length; j++) {
          if (ygrid[j] >= y0-rx && ygrid[j] <= y0+rx) {
            float d2 = (xgrid[i] - x0) * (xgrid[i] - x0) / rx / rx + (ygrid[j] - y0) * (ygrid[j] - y0) / ry / ry;
            if (d2 < 1) {
              for (int r=0; r<animParticleReps; r++) {
                animParticleStart = (Particle[]) append(animParticleStart, new Particle(xgrid[i], ygrid[j], currTime, flows[0]));
              }
            }
          }
        }
      }
    }
    if (animParticleStart.length == 0) {
      // make sure at least one point is included
      float xn = xgrid[findNearest(xgrid,x0)];
      float yn = ygrid[findNearest(ygrid,y0)];
      for (int r=0; r<animParticleReps; r++) {
        animParticleStart = (Particle[]) append(animParticleStart, new Particle(xn, yn, currTime, flows[0]));
      }
    }
  }



  public void startAnimation() { // same for a cold start and a rewind
    animating = true;
    animLastFrameTime = millis() - animFrameInterval;
    animParticles = new Particle[flows.length][animParticleStart.length];
    for (int i=0; i<animParticleStart.length; i++) {
      for (int j=0; j<flows.length; j++) {
        animParticles[j][i] = new Particle(animParticleStart[i].x, animParticleStart[i].y, animParticleStart[i].t, flows[j]);
      }
    }
    currTime = animParticleStart[0].t; // this might have to be replaced with something fancier if the user has their own time controls
  }

  public void stopAnimation() {
    animating = false;
    currTime = 0; // not necessarily the behavior one wants
  }

  public void resumeAnimation() {
    animating = true;
  }
}
// Flow and Particle are the two classes that house the actual technique for representing the flow and tracking particles in it.
// it should be possible to change the internals of the return-map method substantially and not change anything except this file.


class Flow {
  
  boolean visible = true;
  float timestep;
  float[] X, Y;
  float[] T;
  float[][][][] Xnext, Ynext; // time x n_alternates x J x I
  float[][][] mask;
  float[] ind;
  int nAlt, I, J, NF;
  boolean cyclic = true;
  float cyclePeriod;
  String[] filenames;
  boolean indicesInFirstFileOnly = false;
  boolean preloading = true;
  int nextToPreload = 0;
  boolean animatable = true;
  boolean hoverable = true;

  Flow() {}
  
  Flow(String filename) {this(new String[] {filename});}
  
  Flow(String[] filenames) {this(filenames, false);}
  
  Flow(String[] filenames, boolean iiffo) {
    /* the flow data is stored as follows:
    each netcdf file stores one timestep: one displacement map showing where a cartesian grid of points will be after some time interval.
    this map can be multiply valued, since particles might be tracked including random diffusion, such that one start location on the grid
    has multiple end locations. That's what "numAlternates" and "nAlt" are about.
    within each file:
    x, y are 1-d arrays that specify a cartesian grid
    ind is a linear index into points in that grid (mapping the 2 spatial dimensions onto one index the same way matlab does)
    xnext and ynext are arrays the same length (in the first dimension) as ind, giving end locations
    */
    this.filenames = filenames;
    indicesInFirstFileOnly = iiffo;
    NetcdfFile nc = nc_open(filenames[0]);
    X = nc_read1D(nc,"x");
    Y = nc_read1D(nc,"y");
    nAlt = round(nc_readOne(nc,"numAlternates"));
    timestep = nc_readOne(nc,"timestep");
    ind = nc_read1D(nc,"ind"); // ind can theoretically change between files (if there's wetting and drying, for example), but load the first frame's version now in case it doesn't and the other files omit it
    nc_close(nc);
    I = X.length;
    J = Y.length;
    NF = filenames.length;
    Xnext = new float[NF][nAlt][J][I];
    Ynext = new float[NF][nAlt][J][I];
    mask = new float[NF][J][I];
    T = new float[NF];
    preloading = true;    
  }
  
  public void preload() {
    while (preloading) preloadNext();
    preloading = false;
  }
  
  public void preloadNext() {
    if (nextToPreload < NF) {
      // read one frame of flow information
      int fi = nextToPreload;
      NetcdfFile nc = nc_open(filenames[fi]);
      int nCells = (int) nc_readOne(nc,"numCells");
      float[][] xn = new float[nAlt][nCells];
      float[][] yn = new float[nAlt][nCells];
      if (nAlt>1) {
        xn = nc_read2D(nc,"xnext");    
        xn = nc_read2D(nc,"ynext");
      } else {
        xn[0] = nc_read1D(nc,"xnext");
        yn[0] = nc_read1D(nc,"ynext");
      }
      if (!indicesInFirstFileOnly) ind = nc_read1D(nc,"ind");
      if (indicesInFirstFileOnly && nCells != ind.length) println("warning: wrong number of points in file #" + fi + " (ind.length from file #0 = " + ind.length + ", nCells = " + nCells +")");
      nc_close(nc);
      // blank out Xnext, Ynext (setting everything to what a land cells would have)
      for (int r=0; r<nAlt; r++) {
        for (int j=0; j<J; j++) {
          for (int i=0; i<I; i++) {
            Xnext[fi][r][j][i] = X[i];
            Ynext[fi][r][j][i] = Y[j];
            mask[fi][j][i] = 0;
          }
        }
      }
      // now fill in the valid values from xn,yn, rescaling to lat-lon units
      for (int m=0; m<nCells; m++) {
        int i = ind2i((int) ind[m]);
        int j = ind2j((int) ind[m]);
        for (int r=0; r<nAlt; r++) {
          Xnext[fi][r][j][i] = map(xn[r][m], -32768, 32767, X[0], X[I-1]);
          Ynext[fi][r][j][i] = map(yn[r][m], -32768, 32767, Y[0], Y[J-1]); 
          mask[fi][j][i] = 1;
        }
      }
      nextToPreload++;
    } else {
      preloading = false;
    }
  }
  

  public float firstTime() {return T[0];}
  public float lastTime() {return T[T.length-1];}
  public float lastLoadedTime() {return T[max(nextToPreload-1,0)];}
  
  // 1D - 2D index conversions
  public int ij2ind(int i, int j) {return i*Y.length + j;}
  public int ind2i(int ind) {return ind / Y.length;}
  public int ind2j(int ind) {return ind % Y.length;}
  public int[] ind2ij(int ind) {return new int[] {ind % Y.length, ind / Y.length};}
  
  // index - coordinate conversions
  public float i2x(int i0) {return X[round(i0)];}
  public float j2y(int j0) {return Y[round(j0)];}
  public float n2t(int n0) {return T[round(n0)];}
  public int x2i(float x0) {return findNearest(X,x0);}
  public int y2j(float y0) {return findNearest(Y,y0);}  
  public int t2n(float t0) {
    if (cyclic) {
      return findNearest(T,t0 % cyclePeriod);
    } else {
      return findNearest(T,t0);
    }
  }
  public int xy2ind(float x0, float y0) {return ij2ind(x2i(x0), y2j(y0));}
  public float[] ind2xy(int ind) {
    int[] ij = ind2ij(ind);
    return new float[] {i2x(ij[0]), j2y(ij[1])};
  }

  // flow information
  public float[] interpXYnext(float x0, float y0, float t0) {
    int r = floor(random(nAlt));
    float xn = interp2(Y, X, Xnext[t2n(t0)][r], y0, x0);
    float yn = interp2(Y, X, Ynext[t2n(t0)][r], y0, x0);
    return new float[] {xn,yn};
  }
  
  public float interpMask(float x0, float y0, float t0) {
    return interp2(Y, X, mask[t2n(t0)], y0, x0);
  }

}



// -------------------------------------------------------------------------------------



class Particle {
  // a single location/time, updateable but with no memory

  Flow flow;
  float t, age;
  float x, y; // position
  float mask = 1;
  float Kjitter = 10;
  
  Particle() {}
  
  Particle(float x, float y, float t, Flow flow) {
    this.flow = flow;
    this.x = x;
    this.y = y;
    this.t = t;
    mask = flow.interpMask(x,y,t);
    age = 0;
  }
  
  public boolean stuck() { // there are loose and strict ways of defining stuckness
    return (mask < 0.01f);
  }
  
  public void takeStep() {
    t += flow.timestep; // no matter what, time advances
    age += flow.timestep;
    if (!stuck()) {
      // the actual step
      float[] xy1 = flow.interpXYnext(x,y,t);
      x = xy1[0];
      y = xy1[1];
      mask = flow.interpMask(x,y,t);
      // jitter
      float dx = sqrt(2*Kjitter*abs(flow.timestep)) * constrain(randn(),-2,2) / 111325 / cos(y/180.f*PI);
      float dy = sqrt(2*Kjitter*abs(flow.timestep)) * constrain(randn(),-2,2) / 111325;
      float mask1 = flow.interpMask(x+dx,y+dy,t);
      if (mask1 > mask) { // apply the jitter only if it moves particles away from the coastline (or in open water)
        x += dx;
        y += dy;
        mask = mask1;
      }
    }
  }
  
  public void calcToTime(float tend) {
    if (flow.timestep > 0) {
      while (t < tend) takeStep();
    } else {
      while (t > tend) takeStep();
    }
  }
  
  public Particle clone() {
    return new Particle(x,y,t,flow);
  }
  
}


// --------------------------------------------------------------------------------------------------


class Trajectory {// like a particle, but saves its history
  
  Particle[] pos = new Particle[0];
  int length = 0;
  
  Trajectory() {}
  
  Trajectory(float x, float y, float t, Flow flow) {
    addPos(new Particle(x,y,t,flow));
  }
  
  public void addPos(Particle pos1) {
    pos = (Particle[]) append(pos, pos1);
    length++;
  }
  
  public void takeStep() {
    Particle pos1 = pos[length-1].clone();
    pos1.takeStep();
    addPos(pos1);
  }
  
  public void calcToTime(float tend) {
    if (pos[0].flow.timestep > 0) {
      while (pos[length-1].t < tend) takeStep();
    } else {
      while (pos[length-1].t > tend) takeStep();
    }
  }
  
  public void brieflyCalcTowardTime(float tend, float maxCalcTime) {
    float tic = millis();
    if (pos[0].flow.timestep > 0) {   
      while (pos[length-1].t < tend && (millis()-tic < maxCalcTime)) takeStep();
    } else {
      while (pos[length-1].t > tend && (millis()-tic < maxCalcTime)) takeStep();
    }
  }
  
  public float[] x() {
    float[] result = new float[length];
    for (int i=0; i<result.length; i++) result[i] = pos[i].x;
    return result;
  }
  
  public float[] y() {
    float[] result = new float[length];
    for (int i=0; i<result.length; i++) result[i] = pos[i].y;
    return result;
  }
  
  public float[] t() {
    float[] result = new float[length];
    for (int i=0; i<result.length; i++) result[i] = pos[i].t;
    return result;
  }
  
  public float[] age() {
    float[] result = new float[length];
    for (int i=0; i<result.length; i++) result[i] = pos[i].age;
    return result;
  }
  
  public Particle atTime(float t0) {
    float[] T = t();
    if (t0 < T[0] || t0 > T[T.length-1]) return null;
    int n = findNearest(T, t0);
    return pos[n];
  }
 
}
class Gui {
  
  Stylesheet styles = masterStylesheet;
  Display display;
  
  Gui() {}
  
  public void update() {
    if (display==null) return;
    if (display.animating) {
      String timeString = round(display.currTime / 86400.f) + " days";
      styles.setFont(styles.largeFontSize);
      textAlign(RIGHT);
      text(timeString, 10, 10+styles.largeFontSize);
    }
  }
  
  public boolean offerMousePress() {
    return false;
  }
  
  public boolean offerMouseDrag() {
    return false;
  }
  
  public boolean offerMouseRelease() {
    return false;
  }

  
}
// generic_gui 2.1
// neil banas, dec 2010

class Stylesheet { // with the actual stylesheet for the app in the defaults.

  int dkGreenColor = color(10,40,0);
  int medGreenColor = color(66,90,33);
  int ltGreenColor = color(122,140,66);
  int ltltGreenColor = color(169,190,99);
  // ----
  int dkTealColor = color(6,34,44);
  int medTealColor = color(36,72,72);
  int ltTealColor = color(66,110,104);
  int ltltTealColor = color(96,148,136);
  // ----
  int dkBlueColor = color(14,30,60);
  int medBlueColor = color(34,65,105);
  int ltBlueColor = color(54,100,150);
  int ltltBlueColor = color(74,135,195);
  // ----
  int dkPurpleColor = color(57,24,51);
  int medPurpleColor = color(80,53,73);
  int ltPurpleColor = color(103,81,95);
  int ltltPurpleColor = color(126,108,117);
  // ----
  int dkRedColor = color(112,15,0);
  int medRedColor = color(146,54,38);
  int ltRedColor = color(180,90,76);
  int ltltRedColor = color(214,126,114);
  // ----
  int orangeColor = color(186,92,38);
  int yellowColor = color(191,146,54);
  // ----
  int offWhiteColor = color(255,255,243);
  int dkBrownColor = color(77,47,31);
  int medBrownColor = color(125,88,64);
  int ltBrownColor = color(185,142,92);
  int dkGrayColor = color(0);
  int medGrayColor = color(102);
  int ltGrayColor = color(204);
  
  String fontFamily = "BentonSans-Regular";
  int smallFontSize = 10;
  int normalFontSize = 12;
  int largeFontSize = 18;
  int[] fontSizes = {smallFontSize, normalFontSize, largeFontSize};
  PFont[] fonts;
  float guiHeight = 24;
  float guiMargin = 8;
  float textBtnTextSize = 13;
  float iconBtnTextSize = 10;
  float sliderHeight = 10;
  float sliderTextSize = 12;
  int bkgndColor = color(230);
  int labelColor = color(0);
  int[] btnPalette = {color(0), ltGreenColor, medGreenColor, dkGreenColor};
  int sliderLabelColor = color(0);
  int sliderBkgndColor = color(255);
    
  Stylesheet() {}
  
  public void loadFonts() {    
    fonts = new PFont[fontSizes.length];
    for (int i=0; i<fontSizes.length; i++) {
      fonts[i] = loadFont(fontFamily + "-" + fontSizes[i] + ".vlw");
    }
    textFont(fonts[0]);
  }
  
  public int smallFontSize() {return fontSizes[0];}
  public int normalFontSize() {return fontSizes[1];}
  public int largeFontSize() {return fontSizes[2];}
  public int maxFontSize() {return fontSizes[3];}
    
  public void setFont(float fs) {
    fs = constrain(fs, fontSizes[0],fontSizes[fontSizes.length-1]);
    int i=0;
    while (i<fontSizes.length && fs > fontSizes[i]) i++;
    i = constrain(i,0,fontSizes.length-1);
    textFont(fonts[i]);
    textSize(fs);
  }
  
}


// --------------------------------------------------------------------------------------


class GuiElement {
  String name = "";
  float x0, y0, width, height;
  boolean hidden = false;
  boolean alwaysAwake = false;
  Stylesheet styles = masterStylesheet;

  GuiElement() {}
  
  public boolean update() {return false;}  

  public boolean offerMousePress() {return false;}  

  public float[] getPosition() {
    return new float[] {x0, y0, width, height};
  }
  
  public void setPosition(float[] pos) {
    x0 = pos[0];
    y0 = pos[1];
    width = pos[2];
    height = pos[3];
  }
  
  public void setPosition(float x, float y, float wd, float ht) {
    x0 = x;
    y0 = y;
    width = wd;
    height = ht;
  }
  
  public float left() {return x0;}
  public float right() {return x0+width;}
  public float top() {return y0;}
  public float bottom() {return y0+height;}
  
  public boolean over() {
    return mouseX >= left()  &&  mouseX <= right()  &&  mouseY >= top()  &&  mouseY <= bottom();
  }
  
}


// -----------------------------------------------------------------------------------

 
class Button extends GuiElement { // a clickable region, no visible labeling
  boolean awake = false;
  boolean pressed = false;
  
  Button() {}

  Button(String name, float x, float y, float wd, float ht, Stylesheet styles) {
    this.name = name;
    setPosition(x,y,wd,ht);
  }
  
  public int areaColor() {
    if (pressed) {
      return(styles.btnPalette[3]);
    } else if (over() || alwaysAwake) {
      return(styles.btnPalette[2]);
    } else {
      return(styles.btnPalette[1]);
    }
  }
  
  public int textColor() {
    if (pressed) {
      return(styles.btnPalette[3]);
    } else if (over() || alwaysAwake) {
      return(styles.btnPalette[2]);
    } else {
      return(styles.btnPalette[0]);
    }
  }

  public void draw() {
    pushStyle();
    noStroke();
    fill(areaColor());
    rectMode(CORNER);
    rect(x0,y0,width,height);
    popStyle();
  }
  
  public boolean offerMousePress() {
    pressed = over() && (!hidden);
    return pressed;
  }
  
  public boolean update() { // returns true when the button is released.
    boolean result = false;
    if (pressed) {
      awake = over();
      pressed = mousePressed;
    }
    if (awake && (!mousePressed)) {
      pressed = false;
      awake = false;
      result = true;
    } 
    if (!hidden) {
      draw();
    }
    return result;
  }
  
}


// -----------------------------------------------------------------------------------


class TextButton extends Button {
  
  boolean outline = false;
  float textSize;
  
  TextButton(String name, float x, float y, float wd, float ht, Stylesheet styles) {
    this.name = name;
    this.styles = styles;
    textSize = styles.textBtnTextSize;
    setPosition(x,y,wd,ht);
  }

  TextButton(String name, float x, float y, float ht, Stylesheet styles) {
    this.name = name;
    this.styles = styles;
    textSize = styles.textBtnTextSize;
    pushStyle();
    styles.setFont(textSize);
    setPosition(x,y,textWidth("  " + name + "  "),ht);
    popStyle();
  }

  public void draw() {
    pushStyle();
    fill(textColor());
    noStroke();
    styles.setFont(textSize);
    textAlign(CENTER);
    text(name, x0 + width/2.f, y0 + height/2.f + 0.5f*textAscent());   
    if (outline) {
      strokeWeight(1);
      stroke(255);
      noFill();
      rectMode(CORNERS);
      rect(x0, y0, x0+width, y0+height);
    }
    popStyle(); 
  }
  
}


class TextLabel extends TextButton {
  
  TextLabel(String name, float x, float y, float wd, float ht, Stylesheet styles) {
    super(name, x, y, wd, ht, styles);
  }

  TextLabel(String name, float x, float y, float ht, Stylesheet styles) {
    super(name, x, y, ht, styles);
  }
  
  public int textColor() {
    return styles.labelColor;
  }
  
  public boolean update() {
    if (!hidden) draw();
    return false;
  }
    
}


// -----------------------------------------------------------------------------------


class IconButton extends Button {
  
  PImage[] imgs = new PImage[1];
  String[] names = new String[1];
  int current = 0;
  String lastSelected = "";
  
  IconButton(String name, float x, float y, String filename, Stylesheet styles) {
    this.styles = styles;
    names[current] = name;
    imgs[current] = loadImage(filename);
    setPosition(x,y,imgs[current].width,imgs[current].height);
  }
  
  public void addState(String name, String filename) {
    imgs = (PImage[]) append(imgs, loadImage(filename));
    names = (String[]) append(names, name);    
  }

  public void draw() {
    pushStyle();
    if (imgs[current] != null) {
      tint(areaColor());
      image(imgs[current],x0,y0);
    } else {
      fill(areaColor());
      noStroke();
      rectMode(CORNERS);
      rect(x0,y0,x0+width,y0+height);
    }
    fill(textColor());
    noStroke();
    textAlign(CENTER);
    textLeading(1.1f*styles.iconBtnTextSize);
    text(names[current], x0 + width/2.f, y0 + height + 1.5f*textAscent());
    popStyle();
  }
  
  public boolean update() {
    boolean captured = super.update();
    if (captured) {
      lastSelected = names[current];
      current = (current+1) % names.length;
    }
    return captured;
  }

}


// -----------------------------------------------------------------------------------


class Slider extends GuiElement {
  boolean awake = false;
  boolean pressed = false;
  
  float markerWidth;
  boolean showNameOnLeft = true;
  boolean showValOnRight = true;
  float leading = 0.9f;

  float dataMin, dataMax;
  boolean logScale = false;
  float pos = 0.5f; // pos = normalized units
  int decimalPlaces = 2;
  boolean quantized = false;
  float quantizeUnit;
  
  Slider() {}
  
  Slider(String name, float x, float y, float wd, Stylesheet styles, float dataMin, float dataMax) {
    this.name = name;
    this.styles = styles;
    this.dataMin = dataMin;
    this.dataMax = dataMax;
    setPosition(x, y, wd, styles.sliderHeight);
    markerWidth = styles.sliderHeight;
  }
  
  public void quantize(float unit) {
    quantized = true;
    quantizeUnit = unit;
    if (abs(unit-round(unit)) < 1e-6f) decimalPlaces = 0;
    setVal(getVal());
  }

  public int areaColor() {
    if (pressed) {
      return(styles.btnPalette[3]);
    } else if (over() || alwaysAwake) {
      return(styles.btnPalette[2]);
    } else {
      return(styles.btnPalette[1]);
    }
  }

  public void drawMarker() {
    if (pressed) {
      drawMarker(getVal(), styles.btnPalette[3]);
    } else if (over() || alwaysAwake) {
      drawMarker(getVal(), styles.btnPalette[2]);
    } else {
      drawMarker(getVal(), styles.btnPalette[1]);
    }
  }
  
  public void drawMarker(float val, int col) {
    pushStyle();
    rectMode(CORNER);
    noStroke();
    fill(col);
    float x1 = x0 + (width-markerWidth)*val2pos(val);
    rect(x1,y0,markerWidth,height);   
    popStyle(); 
  }
  
  public void writeOnLeft(String S) {
    pushStyle();
    fill(styles.sliderLabelColor);
    textAlign(RIGHT);
    styles.setFont(styles.sliderTextSize);
    textLeading(leading*styles.sliderTextSize);
    text(S+" ",x0,y0+height);
    popStyle();
  }
  
  public void writeOnRight(String S) {
    pushStyle();
    fill(styles.sliderLabelColor);
    textAlign(LEFT);
    styles.setFont(styles.sliderTextSize);
    textLeading(leading*styles.sliderTextSize);
    text(" "+S, x0+width, y0+height);
    popStyle();
  }
  
  public void draw() {
    pushStyle();
    // background
    rectMode(CORNER);
    noStroke();
    fill(styles.sliderBkgndColor);
    rect(x0,y0,width,height);
    // marker
    drawMarker();
    // labeling
    if (showNameOnLeft) writeOnLeft(name);
    if (showValOnRight) writeOnRight(val2string(getVal()));
    popStyle();
  }
  
  public boolean offerMousePress() {
    pressed = over() && (!hidden);
    return pressed;
  }
  
  public boolean update() {
    // returns true if the user is changing the position of the slider.
    boolean result = false;
    if (!hidden) {
      pressed = pressed && mousePressed;
      awake = over() && pressed;
      if (awake) {
        float newpos = (mouseX - x0) / (width - markerWidth);
        if (newpos != getPos()) {
          setPos(newpos);
          result = true;
        }
      }
      draw();
    }
    return result;
  }
  
  // use these four routines for reading & changing the position of the slider.
  // pos = relative position, 0..1
  // val = value in data units  
  public float getPos() {return pos;}
  public float getVal() {return pos2val(getPos());}
  public void setPos(float p) {pos = constrain(p, 0, 1);}
  public void setVal(float v) {setPos(val2pos(v));}

  // these are general conversion functions, that can be used for values other than the current one
  // e.g., to find out the current min and max values allowed, use pos2val(0) and pos2val(1)
  public String val2string(float v) {
    if (decimalPlaces==0) {
      return str(round(v));
    } else {
      float p = pow(10,decimalPlaces);
      return str(round(v*p)/(float)p);
    }
  }
  
  public float val2pos(float v) {
    float p;
    if (quantized) v = round(v/quantizeUnit)*quantizeUnit;
    if (logScale) {
      p = (log(v)-log(dataMin))/(log(dataMax)-log(dataMin));
    } else {
      p = (v-dataMin)/(dataMax-dataMin);
    }
    if (isnan(p)) {
      return 0;
    } else {
      return p;
    }
  }
  
  public float pos2val(float p) {
    float v;
    if (logScale) {
      v = dataMin * pow(dataMax/dataMin, p);
    } else {
      v = dataMin + (dataMax-dataMin) * p;
    }
    if (quantized) v = round(v/quantizeUnit)*quantizeUnit;
    return v;
  }  

}


// -----------------------------------------------------------------------------------


class ImageSlider extends Slider {

  PImage img;
  
  ImageSlider(String name, float x, float y, String filename, Stylesheet styles, float dataMin, float dataMax) {
    this.name = name;
    this.styles = styles;
    this.dataMin = dataMin;
    this.dataMax = dataMax;
    img = loadImage(filename);
    setPosition(x, y, img.width, img.height);
    markerWidth = styles.sliderHeight; // just a guess at something appropriate
    showNameOnLeft = false; // likewise
    showValOnRight = false;
  }
  
  public void draw() {
    image(img, x0, y0);
    drawMarker();
  }

}


// -----------------------------------------------------------------------------------


class Toolbar extends GuiElement {
  
  GuiElement[] elements = new GuiElement[0];
  boolean addFromLeft = true;
  float unoccLeft, unoccRight, margin;
  String lastUpdated;
  
  Toolbar() {}
  
  Toolbar(float x, float y, float wd, Stylesheet styles) {
    this.styles = styles;
    setPosition(x, y, wd, styles.guiHeight);
    margin = styles.guiMargin;
    unoccLeft = x0 + margin;
    unoccRight = x0 + width - margin;
  }

  public TextButton addTextButton(String name, float wd) {
    TextButton btn = new TextButton(name, unoccLeft, y0, wd, height, styles);
    if (addFromLeft) {
      unoccLeft += btn.width + margin;
    } else {
      btn.x0 = unoccRight - btn.width;
      unoccRight = btn.x0 - margin;
    }
    elements = (GuiElement[]) append(elements, btn);
    return btn;
  }

  public TextButton addTextButton(String name) {
    TextButton btn = new TextButton(name, unoccLeft, y0, height, styles);
    if (addFromLeft) {
      unoccLeft += btn.width + margin;
    } else {
      btn.x0 = unoccRight - btn.width;
      unoccRight = btn.x0 - margin;
    }
    elements = (GuiElement[]) append(elements, btn);
    return btn;
  }
  
  public TextLabel addTextLabel(String name, float wd) {
    TextLabel btn = new TextLabel(name, unoccLeft, y0, wd, height, styles);
    if (addFromLeft) {
      unoccLeft += btn.width + margin;
    } else {
      btn.x0 = unoccRight - btn.width;
      unoccRight = btn.x0 - margin;
    }
    elements = (GuiElement[]) append(elements, btn);
    return btn;
  }

  public TextLabel addTextLabel(String name) {
    TextLabel btn = new TextLabel(name, unoccLeft, y0, height, styles);
    if (addFromLeft) {
      unoccLeft += btn.width + margin;
    } else {
      btn.x0 = unoccRight - btn.width;
      unoccRight = btn.x0 - margin;
    }
    elements = (GuiElement[]) append(elements, btn);
    return btn;
  }

  public IconButton addIconButton(String name, String filename) {
    IconButton btn = new IconButton(name, unoccLeft, y0, filename, styles);
    if (addFromLeft) {
      unoccLeft += btn.width + margin;
    } else {
      btn.x0 = unoccRight - btn.width;
      unoccRight = btn.x0 - margin;
    }
    elements = (GuiElement[]) append(elements, btn);
    return btn;
  }

  public Slider addSlider(String name, float wd, float dataMin, float dataMax) {
    Slider sl = new Slider(name, unoccLeft, y0 + (height-styles.sliderHeight)/2.f, wd, styles, dataMin, dataMax);
    pushStyle();
    styles.setFont(styles.sliderTextSize);
    float leftTextSpace = textWidth(sl.name+" ");
    float rightTextSpace = max(textWidth(" "+sl.val2string(dataMin)), textWidth(" "+sl.val2string(dataMax)));
    popStyle();
    if (addFromLeft) {
      unoccLeft += leftTextSpace + sl.width + rightTextSpace + margin;
    } else {
      sl.x0 = unoccRight - sl.width - rightTextSpace;
      unoccRight = sl.x0 - leftTextSpace - margin;
    }
    elements = (GuiElement[]) append(elements, sl);
    return sl;
  }
  
  public void addSpacer() {
    addSpacer(height);
  }
  
  public void addSpacer(float wd) {
    if (addFromLeft) {
      unoccLeft += wd + margin;
    } else {
      unoccRight -= wd + margin;
    }
  }
  
  public void select(String key) {
    for (int i=0; i<elements.length; i++) {
      if (elements[i].name.equals(key)) {
        elements[i].alwaysAwake = true;
      }
    }
  }
  
  public void deselect(String key) {
    for (int i=0; i<elements.length; i++) {
      if (elements[i].name.equals(key)) {
        elements[i].alwaysAwake = false;
      }
    }
  }

  public void deselect() {
    for (int i=0; i<elements.length; i++) {
      elements[i].alwaysAwake = false;
    }
  }
  
  public boolean update() {
    boolean captured = false, result = false;
    int i;
    for (i=0; i<elements.length; i++) {
      captured = elements[i].update();
      if (captured) {
        lastUpdated = elements[i].name;
        result = true;
      }
    }
    return result;
  }
  
  public boolean offerMousePress() {
    boolean captured = false;
    for (int i=0; i<elements.length && !captured; i++) {
      captured = elements[i].offerMousePress();
    }
    return captured;
  }
  
}


// ---------------------------------------------------------------------------------


class SliderList extends Toolbar {

  SliderList(float x, float y, float wd, Stylesheet styles) {
    this.styles = styles;
    setPosition(x, y, wd, 0);
    margin = styles.guiMargin;
  }
  
  public Slider addSlider(String name, float dataMin, float dataMax) {
    Slider sl = new Slider(name, x0, y0+height, width, styles, dataMin, dataMax);
    elements = (GuiElement[]) append(elements, sl);
    height += styles.sliderHeight + margin;
    return sl;
  }
  
  public void addSpacer() {
    height += margin;
  }

}






// opening and closing -----------------------------------------------


public NetcdfFile nc_open(String filename) {
  NetcdfFile nc = null;
  try {
    nc = NetcdfFile.open(filename);
  } catch (IOException ioe) {
    println("error opening " + filename + ": " + ioe.toString());
  }
  return nc;
}


public void nc_close(NetcdfFile nc) {
  if (nc==null) return;
  try {
    nc.close();
  } catch (IOException ioe) {
    println(ioe.toString());
  }
}


public Variable nc_getVar(NetcdfFile nc, String varname) {
  if (nc==null) return null;
  Variable v = nc.findVariable(varname);
  if (v==null) {
//    println(varname + " not found");
    return null;
  } else {
    return v;
  }
}



// reading entire variables -----------------------------------------



public float nc_readOne(NetcdfFile nc, String varname) {
  float result = 0.f/0;
  Variable v = nc_getVar(nc, varname);
  if (v != null) {
    try {
      result = v.readScalarFloat();      
    } catch (Exception e) {
      println(e.toString());
    } 
  }
  return result;
}



public float[] nc_read1D(NetcdfFile nc, String varname) {
  float[] result = null;
  Variable v = nc_getVar(nc, varname);
  if (v != null) {
    try {
      Array a = v.read().reduce(); 
      int[] s = a.getShape();
      result = new float[s[0]];
      Index ind = a.getIndex();
      for (int i=0; i<s[0]; i++) {
        result[i] = a.getFloat(ind.set(i));
      }
    } catch (Exception e) {
      println(e.toString());
    } 
  }
  return result;
}

  
public float[][] nc_read2D(NetcdfFile nc, String varname) {
  float[][] result = null;
  Variable v = nc_getVar(nc, varname);
  if (v != null) {
    try {
      Array a = v.read().reduce(); 
      int[] s = a.getShape();
      result = new float[s[0]][s[1]];
      Index ind = a.getIndex();
      for (int j=0; j<s[0]; j++) {
        for (int i=0; i<s[1]; i++) {
          result[j][i] = a.getFloat(ind.set(j,i));
        }
      }
    } catch (Exception e) {
      println(e.toString());
    } 
  }
  return result;
}


public float[][][] nc_read3D(NetcdfFile nc, String varname) {
  float[][][] result = null;
  Variable v = nc_getVar(nc, varname);
  if (v != null) {
    try {
      Array a = v.read().reduce(); 
      int[] s = a.getShape();
      result = new float[s[0]][s[1]][s[2]];
      Index ind = a.getIndex();
      for (int k=0; k<s[0]; k++) {
        for (int j=0; j<s[1]; j++) {
          for (int i=0; i<s[2]; i++) {
            result[k][j][i] = a.getFloat(ind.set(k,j,i));
          }
        }
      }
    } catch (Exception e) {
      println(e.toString());
    } 
  }
  return result;
}
// color -----------------------------------------------------------------------------------------

public int shift(int col, float r) {
  if (r>0) {
    return lerpColor(col,color(255,255,255,alpha(col)),r);
  } 
  else {
    return lerpColor(col,color(0,0,0,alpha(col)),-r);
  }
}

public int lighten(int col) {
  return shift(col,0.25f);
}

public int darken(int col) {
  return shift(col,-0.25f);
}

public int transparency(int col, float f) {
  return color(red(col),green(col),blue(col),f*alpha(col));
}

public int desaturate(int col, float f) {
  return lerpColor(col, desaturate(col), f);
}
public int desaturate(int col) {
  return color(brightness(col),brightness(col),brightness(col),alpha(col));
}

int dkBrownColor = color(77,47,31);
int medBrownColor = color(125,88,64);
int ltBrownColor = color(185,142,92);
int medGreenColor = color(66,90,33);
int ltGreenColor = color(122,140,66);
int ltltGreenColor = color(169,190,99);
int dkBlueColor = color(14,30,60);
int medBlueColor = color(34,65,105);
int ltBlueColor = color(54,100,150);
int ltltBlueColor = color(74,135,195);
int dkPurpleColor = color(57,24,51);
int medPurpleColor = color(80,53,73);
int dkRedColor = color(112,15,0);
int medRedColor = color(146,54,38);
int ltRedColor = color(180,90,76);
int ltltRedColor = color(214,126,114);
int orangeColor = color(186,92,38);
int yellowColor = color(191,146,54);
int offWhiteColor = color(255,255,243);
int dkGrayColor = color(0);
int medGrayColor = color(102);
int ltGrayColor = color(204);



// math -----------------------------------------------------------------------------------------


public float[] firstRow(float[][] A) {
  return A[0];
}

public float[] firstCol(float[][] A) {
  float[] x = new float[A[0].length];
  for (int i=0; i<x.length; i++) x[i] = A[i][0];
  return x;
}


public float[] zeros(int I) {
  float[] a = new float[I];
  for (int i=0; i<I; i++) a[i] = 0;
  return a;
}
public float[][] zeros(int J, int I) {
  float[][] a = new float[J][I];
  for (int j=0; j<J; j++) for (int i=0; i<I; i++) a[j][i] = 0;
  return a;
}
public float[][][] zeros(int K, int J, int I) {
  float[][][] a = new float[K][J][I];
  for (int k=0; k<K; k++) for (int j=0; j<J; j++) for (int i=0; i<I; i++) a[k][j][i] = 0;
  return a;
}


float Inf = 1.f/0.f;
float almostInf = 1e20f;
public boolean isfinite(float a) {
  return (a<almostInf) && (a>-almostInf);
}
public boolean isnan(float a) {
  return !((a>0) || (a<=0));
}


public void finitize(float[] a) {
  for (int i=0; i<a.length; i++) {
    if (!isfinite(a[i])) a[i] = 0;
  }
}
public void finitize(float[][] a) {
  for (int j=0; j<a.length; j++) {
    for (int i=0; i<a[0].length; i++) {
      if (!isfinite(a[j][i])) a[j][i] = 0;
    }
  }
}
public void finitize(float[][][] a) {
  for (int k=0; k<a.length; k++) {
    for (int j=0; j<a[0].length; j++) {
      for (int i=0; i<a[0][0].length; i++) {
        if (!isfinite(a[k][j][i])) a[k][j][i] = 0;
      }
    }
  }
}


public int findIndexBefore(float[] x, float xi) {
  // assumes x is monotonic and increasing
  // return -1 if xi is out of range
  if (xi < x[0]) return -1;
  if (xi > x[x.length-1]) return -1;
  int nbefore = 0;
  int nafter = x.length-1;
  while (nafter-nbefore > 1) {
    int nmid = (nbefore+nafter)/2;
    if (xi > x[nmid]) {
      nbefore = nmid;
    } 
    else {
      nafter = nmid;
    }
  }
  return nbefore;
}


public int findNearest(float[] x, float xi) {
  // assumes x is monotonic and increasing
  // always returns a valid index
  if (xi <= x[0]) return 0;
  if (xi >= x[x.length-1]) return x.length-1;
  int nbefore = 0;
  int nafter = x.length-1;
  while (nafter-nbefore > 1) {
    int nmid = (nbefore+nafter)/2;
    if (xi > x[nmid]) {
      nbefore = nmid;
    } 
    else {
      nafter = nmid;
    }
  }
  if (xi < 0.5f*(x[nbefore]+x[nafter])) {
    return nbefore;
  } 
  else {
    return nafter;
  }
}


public float interp2(float[] y, float[] x, float[][] z, float yi, float xi) {
  int j0 = findIndexBefore(y,yi);
  int i0 = findIndexBefore(x,xi);
  if (i0==-1 || j0==-1) return z[findNearest(y,yi)][findNearest(x,xi)];
  float b = (yi-y[j0]) / (y[j0+1]-y[j0]);
  float a = (xi-x[i0]) / (x[i0+1]-x[i0]);
  float result = (1-b)*(1-a)*z[j0][i0] + b*(1-a)*z[j0+1][i0] + (1-b)*a*z[j0][i0+1] + b*a*z[j0+1][i0+1];
  return result;
}


public float randn() { // normally distributed random number with std dev 1
  return sqrt(-2*log(random(1)))*cos(TWO_PI*random(1));
}


public float[] centerdiff(float[] x) {
  float[] dx = new float[x.length];
  for (int i=1; i<x.length-1; i++) {
    dx[i] = 0.5f*(x[i+1]-x[i-1]);
  }
  dx[0] = dx[1];
  dx[x.length-1] = dx[x.length-2];
  return dx;
}



// strings -----------------------------------------------------------------------------------------



public String fileSuffix(String filename) {
  String[] tok = split(filename,'.');
  if (tok.length<2) {
    return null;
  } 
  else {
    return tok[tok.length-1];
  }
}


public String nicenum(float a) {
  if (a==round(a)) {
    return "" + round(a);
  } else {
    return "" + a;
  }
}


public void textVAlign(String S, float x, float y) {textVAlign(S,x,y,"baseline");}
public void textVAlign(String S, float x, float y, String align) {
  if (align.equals("baseline")) {
    text(S,x,y);
  } else if (align.equals("middle")) {
    text(S,x,y+0.5f*(textAscent()));
  } else if (align.equals("bottom")) {
    text(S,x,y-textDescent());
  } else if (align.equals("top")) {
    text(S,x,y+textAscent());
  }
}
  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#c0c0c0", "flowWeaver" });
  }
}
